_ENV = AIContext

CombatAI_Method = CombatAI_Method or {}

function CombatAI_Method.init_turn()
  local new_turn = setmetatable({}, {__index = CombatAI_Method})

  -- need to get an object belonging to the current side for enemies()
  -- would be nice if enemies() accepted current_side() as an argument...
  for obj in objects() do
    if obj.empire == current_side()
      new_turn.friendly_obj = obj
      break
    end
  end

  new_turn:evaluate_boarding_power()

  -- lazy-populate priority caches
  new_turn.ship_priority = setmetatable({}, {
      __index = function(t, ship)
        new_turn:update_ship_priority(ship)
        return t[ship]
      end
    })
  new_turn.boarding_priority = setmetatable({}, {
      __index = function(t, ship)
        new_turn:update_boarding_priority(ship)
        return t[ship]
      end
    })
  new_turn.combat_power = setmetatable({}, {
      __index = function(t, obj)
        if is_ship(obj)
          rawset(t, obj, new_turn:ship_combat_power(obj))
        elseif is_tracking_object(obj)
          rawset(t, obj, new_turn:tracker_combat_power(obj))
        end
        return t[obj]
      end
    })

  -- find the priority of the most valuable friendly ship
  -- used as a hypothetical target for hostile launchers
  -- find the priority of the most valuable/vulnerable friendly ship
  local max_priority = 0
  for ship in current_sides_ships() do
    max_priority = max(max_priority, new_turn.ship_priority[ship])
  end
  new_turn.highest_friendly_priority = max_priority

  new_turn.estimated_tracker_survival = evaluate_hostile_point_defence(new_turn)

  new_turn:evaluate_fleet_power()
  new_turn:build_target_table()
  new_turn:build_tracker_tables() --NOTE that tracker lists may contain dead objects, due to autofire
  new_turn:build_launcher_tables()

  -- table mapping objects => list of friendly pd/interceptors reserved for that object
  new_turn.reserved_weapons = {}

  return new_turn
end

function CombatAI_Method:update_ship_priority(target)
  self.ship_priority[target] = self:base_ship_priority(target)
end

function CombatAI_Method:update_boarding_priority(target)
  self.boarding_priority[target] = boarding_priority(target)
end

-- check if obj is enemies with the current side
-- the only reason this exists is because enemies() cannot take the current_side()
-- as an argument, so we need to find some friendly object to compare with.
-- We can't use obj.empire == current_side() because that fails for battles involving allies
-- If enemies() is ever updated to be able to take an empire as an argument, 
-- just replace this with enemies(obj, current_side()) everywhere.
function CombatAI_Method:is_hostile(obj)
  if(not self.friendly_obj)
    return true
  end
  return enemies(self.friendly_obj, obj)
end

----- Target Table

function CombatAI_Method:build_target_table()

  --i.e. hostile ships
  self.hostile_targets = {}

  -- attack enemy combat ships
  for target in enemy_combat_ships() do
    self.hostile_targets[target] = true
  end

  -- if there are no combat ships left, then attack all enemy ships
  if(not next(self.hostile_targets))
    for target in current_enemy_ships() do
      self.hostile_targets[target] = true
    end
  end
end

----- Tracker Tables

function CombatAI_Method:build_tracker_tables()

  -- table of hostile trackers
  -- NOTE since autofire can kill trackers during our turn and we never know about it
  -- this table can contain dead trackers and that should be checked for when iterating
  self.hostile_trackers = {}

  -- table mapping objects => list of trackers targeting that object
  -- see register_enroute_tracker()
  self.enroute_trackers = {}

  -- populate tables
  for obj in objects() do
    if is_tracking_object(obj)
      if self:is_hostile(obj)
        self.hostile_trackers[obj] = true
      end
      if enemies(obj, obj.dest)
        self:register_enroute_tracker(obj, obj.dest)
      end
    end
  end

end

----- Launcher Tables

function CombatAI_Method:build_launcher_tables()
  -- table of hostile launchers
  -- used when deciding to reserve interceptors for dangerous trackers not yet launched
  self.hostile_launchers = {}
  self.hostile_shuttles = {} --because assault shuttle bays are not weapons :(

  for ship in current_enemy_ships() do
    for w in weapons(ship)
      if w.is_launcher and alive_weapons(w) > 0
        self.hostile_launchers[w] = true
      end
    end

    --shuttles are not launchers, unfortunately
    local is_boarder, sys = is_combat_boarder(ship)
    if is_boarder and sys=='Assault Shuttles'
      self.hostile_shuttles[ship] = true
    end
  end
end


function CombatAI_Method:drop_target(target)
  --remove target from the target table
  self.hostile_targets[target] = nil

  --also drop any launchers attached to that target
  self.hostile_shuttles[target] = nil
  for launcher ; self.hostile_launchers do
    if launcher.object == target
      self.hostile_launchers[launcher] = nil
    end
  end
end

--register a tracker in flight
function CombatAI_Method:register_enroute_tracker(tracker, target)
  local tracker_data = { damage_per=average_tracker_damage(tracker), n=tracker.n }

  if(not self.enroute_trackers[target])
    self.enroute_trackers[target] = { tracker_data }
  else
    table.insert(self.enroute_trackers[target], tracker_data)
  end
end

--register a tracker launch. make sure this is called before launcher actually fires.
function CombatAI_Method:register_tracker_launch(launcher, target)
  local tracker_data = { damage_per=average_weapon_damage(launcher), n=launcher:potential_shots() }

  if(not self.enroute_trackers[target])
    self.enroute_trackers[target] = { tracker_data }
  else
    table.insert(self.enroute_trackers[target], tracker_data)
  end
end

--register that a weapon is reserved for a potential unlaunched or autofire target
function CombatAI_Method:register_reserved_weapon(reserved_weapon, object)
  local reserved_data = { damage_per=average_weapon_damage(reserved_weapon), n=useable_weapons(reserved_weapon) }

  if(not self.reserved_weapons[object])
    self.reserved_weapons[object] = { reserved_data }
  else
    table.insert(self.reserved_weapons[object], reserved_data)
  end

  --print("reserving", reserved_weapon.object.name, reserved_weapon.part, "for", launcher.object.name, launcher.part, "damage_per", reserved_data.damage_per, "n", reserved_data.n, "#", #self.reserved_weapons[launcher])
end
