_ENV = AIContext

depends '@CombatAIMethods'
depends 'combat_helpers'
depends 'attack_priorities'
depends 'boarding'
depends 'retreat'

-- local we have to duplicate
local function nearest(object_iter,x,y)
  local function length(dx,dy)
    return sqrt(dx*dx+dy*dy)
  end

  local function length_to(a, b)
    ax, ay = object_center(a)
    bx, by = object_center(b)
    return length(ax-bx, ay-by)
  end

  -- local we have to duplicate
  local function object_point_dist(object,x1,y1)
    local x0,y0=object_center(object)
    return length(x0-x1,y0-y1)
  end

  local rval
  local dist
  for obj in object_iter() do
    if(not dist)
      dist=object_point_dist(obj,x,y)
      rval=obj
    else
      local mydist=object_point_dist(obj,x,y)
      if(mydist<dist)
        dist=mydist
        rval=obj
      end
    end
  end

  return rval
end

-- returns an estimate of the distance a ship might move in a single turn
local function estimated_move_dist(ship)
  if ship.cant_move
    return 0
  end

  --as a fudge factor, assume a 45 degree turn
  local turn_cost
  if ship.s < 4
    turn_cost = 25/4
  elseif ship.s == 4
    turn_cost = 25/4*0.7
  else
    turn_cost = 25/4/3
  end
  return max(ship.max_fuel - turn_cost, 0)
end

-- for debugging
local function print_target_table(ship, weapon, target_table, n)
  print(string.format("TARGETS TABLE for %s, %s", ship.name, weapon.part))
  for i, data in ipairs(target_table) do
    local name
    if data.target.is_weapon
      name = string.format("%s.%s", data.target.object.name, data.target.part)
    else
      name = data.target.name
    end
    print(string.format("%s : %.3f %s", name, data.priority, data.reserve and "(RESERVE)" or ""))
    if(n and i >= n) break end
  end
end

-- The core weapon targeting routine
function CombatAI_Method:attack_as_able(ship, temp_reserve_list)
  local to_target in range_fun

  local weapons = ship.weapons
  if not weapons
    return false
  end

  local fire_map={}

  -- populate fire_map
  local j=1
  while (weapons[j]) do
    local w=weapons[j]
    j=j+1

    if temp_reserve_list?[w]
      goto continue --reserved
    end

    local n = w:potential_shots()

    if(n<=0) goto continue end

    -- compile a table of targets by priority
    local target_table = {}


    if w.is_interceptor

      -- consider reserving interceptors for trackers not yet launched
      for launcher ; self.hostile_launchers do
        target_table[#target_table+1] = { reserve=true, target=launcher, object=launcher.object, priority=self:get_launcher_priority(launcher, w) }
      end

    elseif w.is_pd

      -- consider reserving PD for trackers not yet launched, but only if they are within launch distance of us
      for launcher ; self.hostile_launchers do
        local object in launcher
        local dist = to_target(object, ship)
        local launcher_range = launcher:launcher_range() + estimated_move_dist(object)
        if dist < launcher_range
          --since we are only considering local defence here, use our own priority instead of the highest friendly priority
          local local_defence_priority = (self.ship_priority[ship] + self.highest_friendly_priority)/2
          target_table[#target_table+1] = { reserve=true, target=launcher, object=object, priority=self:get_launcher_priority(launcher, w, local_defence_priority) }
        end
      end

      -- consider reserving PD for assault shuttles
      for boarder ; self.hostile_shuttles do
        local dist = to_target(boarder, ship)
        local boarding_range = boarder:boarding_range() + estimated_move_dist(boarder)
        if(0 < dist and dist < boarding_range)
          target_table[#target_table+1] = { reserve=true, target=boarder, object=boarder, priority=self:get_assault_shuttle_priority(boarder, w) }
        end
      end

    end

    -- consider targeting trackers
    if(not w.cant_attack_trackers)
      for tracker ; self.hostile_trackers do
        if not tracker.dead and w:can_attack(tracker)
          target_table[#target_table+1] = { target=tracker, object=tracker, priority=self:get_tracker_priority(tracker, w) }
        end
      end
    end

    -- consider targeting ships
    for target ; self.hostile_targets do
      if w:can_attack(target)
        target_table[#target_table+1] = { target=target, object=target, priority=self:get_adjusted_ship_priority(target, w) }
      end
    end

    if not next(target_table)
      goto continue
    end

    table.sort(target_table,|) function(a,b)
      -- doing something takes precedence over doing nothing, in the case of ties
      -- prevents possible stalemate when all non-PD weapons have been destroyed
      if a.priority == b.priority
        return (not a.reserve) and b.reserve
      end
      return a.priority > b.priority
    end

    --if ship.name == "Heavy Cruiser"
    --  print_target_table(ship, w, target_table)
    --end

    -- choose a target from the target table
    for i, item in ipairs(target_table) do
      local target, object, reserve in target_table[i]

      --TODO need to find a way to figure out if there are any enemy players that move after us this turn
      -- if we are bugging out at the end of this turn, don't bother reserving anything
      --if reserve and ship.retreating and ship.retreating <= 1
      --  goto next_target
      --end

      -- if hostile PD is intense, limit our attacks with fightercraft to targets 
      -- within a certain distance in order to preserve our craft
      if w.is_fighter and self.estimated_tracker_survival < 1
        local max_sortie_dist = 0.45*w:launcher_range()/(1-self.estimated_tracker_survival)
        if reserve and is_ship(object)
          max_sortie_dist += estimated_move_dist(object)
        end
        if(to_target(ship, object) > max_sortie_dist)
          goto next_target
        end
      end

      if reserve
        -- since craft don't autofire, if we are reserving and there are any targets in roundtrip range, 
        -- attack those instead. We'll still be able to launch at our reserve target next turn
        if w.is_fighter
          local roundtrip_range = 0.45*w:launcher_range()
          for i, secondary in ipairs(target_table) do
            if(not secondary.reserve and to_target(ship, secondary.object) < roundtrip_range)
              fire_map[#fire_map+1] = {..w,..n, target=secondary.target }
              break
            end
          end
        end

        -- weapon reserved for this turn
        temp_reserve_list?[w] = true
        self:register_reserved_weapon(w, target)
        break

      else

        -- weapon assigned to target
        fire_map[#fire_map+1] = {..w,..n,..target}
        break

      end

      ::next_target::
    end

    ::continue::
  end


  -- fire weapons in batches based on whether or not they're
  -- weapon_group_equiv.
  local unfired = 0
  if fire_map[1]

    if ship.cloaked
      action.decloak {ship}
    end

    table.sort(fire_map,|) function(a,b)
        return weapon_group_order(a.w,b.w)
    end

    j=1
    for i=1,100 
      if not fire_map[j] break end
      local w,target in fire_map[j]
      local push, sub_fire = pusher {}

      while fire_map[j] and weapon_group_equiv(w,fire_map[j].w) and target==fire_map[j].target
        push | fire_map[j]
        j++
      end

      -- register the launch of any new trackers
      local w, target in sub_fire[1]
      if w.is_launcher
        self:register_tracker_launch(w, target) --make sure this is called before w is fired!
      end
      event_sep_scope(.35,action.fire_weapons, unpack(sub_fire))

      -- after attacking, update the target's priority to reflect possible battle damage
      if is_ship(target)
        if target.dead
          self:drop_target(target)
        else
          self:update_ship_priority(target)
          self:update_boarding_priority(target)
        end
      end
    end
    
  end


  return #fire_map>0
end

-- chooses what to move towards
function CombatAI_Method:choose_move_target(ship, desired_attack_dist)
  local to_target in range_fun

  --iterate all hostile ships and trackers
  local iter_targets = function()
    return chain(self.hostile_targets, self.hostile_trackers)
  end

  local closest = closest_objects(iter_targets, object_center(ship))

  -- choose the best move target out of the closest three ships.
  -- testing has shown we don't want the AI to be too 'clever' with how it chooses
  -- a move target here. Moving the fleet forwards as a group is more important than
  -- closing in on the 'best' target. 
  -- However, some flexibility is beneficial so we pick the best target out of the closest 3 ships
  local best_target
  local best_priority

  local ship_count = 0
  for i = 1,#closest do
    local target = closest[i]
    local target_dist = to_target(ship, target)

    if is_ship(target)
      if ship_count > 3
        break
      end
    end

    --find the best target
    local priority = 0
    if is_combat_boarder(ship)
      if is_ship(target)
        priority = self.boarding_priority[target] or 0
      end
    else
      if is_tracking_object(target)
        priority = self:base_tracker_priority_each(target) * target.n
      --planets have lowest move priority unless they are already within ~1 move of our desired attack distance
      elseif not (target.is_planet and target_dist > desired_attack_dist + estimated_move_dist(ship) + 4)
        priority = self.ship_priority[target]
      end
    end

    if(not best_target or (priority > best_priority))
      best_target = target
      best_priority = priority
    end
  end

  return best_target
end

function CombatAI_Method:close_and_attack(ship)
  local to_target in range_fun

  local move_target

  local attack_dist = desired_attack_dist(ship, self.estimated_tracker_survival)
  if(not attack_dist)
    return
  end

  -- if our move target dies, choose a new one and repeat
  sanity = 0
  while(not ship.dead and (not move_target or move_target.dead)) do
    sanity++
    if sanity > 100
      break 
    end

    if(not move_target or move_target.dead or move_target.defenseless or not enemies(ship,move_target))
      move_target = self:choose_move_target(ship, attack_dist) or closest_enemy_ship(object_center(ship))
    end

    -- if there are no valid targets left, exit the loop
    if(not move_target)
      break
    end

    -- handle movement
    if ship.fuel > 0 and not ship.cant_move and not stand_off_and_bombard

      if stand_off_and_bombard
        attack_dist = max(attack_dist, ranges.long*1.25)
      end

      -- move and attack the current target

      local x,y,r,r2 = find_min_facing_move_L2(ship,move_target,attack_dist)

      if(x)

        if to_target(ship, move_target) <= attack_dist
          --print(ship.name, "NOT MOVING, WITHIN ATTACK DISTANCE", to_target(ship, move_target), "<=", attack_dist)
          x,y = ship.x, ship.y --hack to ensure ships turn in place without moving, causes them to wiggle but that's better than what it was doing before
        end

        action.move_ships{ship, x, y, r}

        if(ship.dead) break end

        if(r2 and r2~=r and ship_can_rotate(ship,r2))
          if(ship.speed==0)
            error("that's very strange")
          end

          action.move_ships{ship, x, y, r2}
        end
      end
    end

    -- in case autofire killed us
    if(ship.dead) break end

    -- fire weapons
    local attacked_this_turn = false
    --keep calling attack_as_able until we run out of weapons
    local temp_reserve_list = {}
    while (self:attack_as_able(ship, temp_reserve_list)) do
      attacked_this_turn = true
      if(move_target.dead) break end
    end

    -- set a new target, if this one has been destroyed
    if(move_target.dead)
      if is_ship(target)
        self:drop_target(move_target)
      end
      goto continue
    end

    if not ship.cant_electrocyte
      try_use_electrocyte(ship,move_target)
    end

    -- combat boarders take more effort to choose a boarding target
    if(not ship.has_boarded and is_combat_boarder(ship))
      local boarding_targets = {}
      for target ; self.hostile_targets
        if can_board(ship,target)
          boarding_targets[#boarding_targets+1] = { target=target, priority=self.boarding_priority[target] }
        end
      end

      table.sort(boarding_targets,|) function(a,b)
        return a.priority > b.priority
      end

      for i,item in ipairs(boarding_targets) do
        if(not self:capture_if_able(ship, item.target)) break end
      end
    end

    -- opportunistic boarding by non-combat-boarders
    if(not move_target.dead and not ship.has_boarded)
      -- don't attempt to board in the thick of combat, unless we have a boarding module
      if(not attacked_this_turn or all_enemies_are_immobile())
        if(self:capture_if_able(ship,move_target))
          self:capture_closest(ship)
        end
      end
    end

    ::continue::
  end

end


-- return the current side's ships in the order that the AI should move them
local function current_side_move_order(ai_data)
  local enroute_trackers in ai_data

  local sort_ships = {}
  for ship in current_sides_ships() do
    sort_ships[#sort_ships+1] = { 
      ship = ship, 
      combat_power = ai_data.combat_power[ship], 
      is_combat_boarder = is_combat_boarder(ship) and -1 or 1, 
      interception_warning = (enroute_trackers[ship] and #enroute_trackers[ship] > 0) and -1 or 1,
    }
    --local ship, is_combat_boarder, interception_warning in sort_ships[#sort_ships]
    --print(ship.name, is_combat_boarder, interception_warning)
  end

  table.sort(sort_ships,|) function (a,b)
    -- first sort by interception warning, ships that have incoming trackers move last
    if a.interception_warning ~= b.interception_warning
      return a.interception_warning > b.interception_warning
    end

    -- then non combat boarders move before combat boarders
    if a.is_combat_boarder ~= b.is_combat_boarder
      return a.is_combat_boarder > b.is_combat_boarder
    end

    -- then by combat value - 
    -- ships with high attack value are more likely to affect the decisions of other ships
    -- and ships with high defence should try to move sooner to keep at the front
    return a.combat_power > b.combat_power
  end

  local ships = {}
  for i=1,#sort_ships do
    ships[#ships+1] = sort_ships[i].ship
  end

  return ships
end

-- for debugging. prints out a summary of the given side's forces, 
-- including their base ship/tracker priority
local function print_force_summary(ai_data, empire)
  print("SHIPS FOR SIDE", empire.name)

  local fleet_power = 0

  print("ship : att, def, combat power : priority")
  print("----------------------------------------")
  for obj in objects() do
    if obj.empire == empire
      -- ships
      if is_ship(obj)
        local attack_value = calc_attack_value(obj) + ai_data:boarding_attack_value(obj)
        local defence_value = calc_defence_value(obj)
        local combat_power = ai_data.combat_power[obj]
        fleet_power += combat_power
        printf("%s : %.1f, %.1f, %.1f : %.3f", obj.name, attack_value, defence_value, combat_power, ai_data.ship_priority[obj])
      -- fightercraft
      elseif is_tracking_object(obj) and obj.is_fighter
        local damage = average_tracker_damage(obj)
        local combat_power = ai_data.combat_power[obj]
        fleet_power += combat_power
        printf("%s : (%.1f, %.1f) x %d, %.1f : %.3f", obj.name, damage, obj.health_per, obj.n, combat_power, ai_data:base_tracker_priority_each(obj) * obj.n )
      end
    end
  end

  print("----------------------------------------")
  print("TOTAL FLEET POWER", math.round(fleet_power))
end

local function inside_do_turn()
  if battle_is_over()
    return
  end

  local ai_turn = CombatAI_Method.init_turn()
  --print_force_summary(ai_turn, current_side())

  local try_retreat_all
  if get_ai_hint() =='retreat_all' or ai_turn.outmatched > ai_turn:general_retreat_threshold()
    console("ordering a general retreat", ai_turn.outmatched)
    try_retreat_all = general_retreat()
  end

  if not try_retreat_all
    --limit bombardment to the first 20 turns
    if all_enemies_are_immobile() and bombardment_available() and ai_turn.estimated_tracker_survival > 0.25 and current_side_nturns() <= 20
      print('standing off')
      stand_off_and_bombard = true
    else
      stand_off_and_bombard = false
    end
  end

  local ships = current_side_move_order(ai_turn)

  for i,ship in ipairs(ships) do
    if( battle_is_over() )
      sprint("looks like the battle is over")
      break
    end

    if(ship.dead)
      goto continue
    end

    if ai_turn:wants_to_retreat(ship, try_retreat_all)
      ai_turn:do_retreat(ship)
    else
      ai_turn:close_and_attack(ship)
    end

    ::continue::
  end

  if (last_event_command()=="end_turn(  )"  and not stand_off_and_bombard) or current_side_nturns() > 40

    console("should probably offer a cease fire", last_event_command(), stand_off_and_bombard, current_side_nturns() )
    local empire = current_side()
    local attack = all_attacks[get_attack_id()]

    for other_empire in pairs(attack.sides) 
      if Battle.enemy(empire,other_empire) and not Battle.get_cease_fire_offer(empire,other_empire)
        console("offering a cease fire to ",other_empire,empire)
        action.offer_cease_fire(empire,other_empire)
      end

      if empire~=other_empire
        print('cease1', empire, other_empire, objects.cease_fire_offers[ 
          empire._ikey ..'_'.. other_empire._ikey ],
          Battle.get_cease_fire_offer(empire,other_empire)
        )

        print('cease2', other_empire.name, empire.name, objects.cease_fire_offers[ 
          other_empire._ikey ..'_'.. empire._ikey ],
          Battle.get_cease_fire_offer(other_empire,empire)
        )
      end
    end
  end
end


function close_and_attack_turn()
  --console("ai turn for", current_side().name, current_side(), current_side_nturns())

  --print('starting', 'temp AI State', v2s | check_var("temp AI State"))
--  console(last_event_command())

  local ok, err = pcall(inside_do_turn)
  if (not ok) and err
    add_exception(tostring | err)
  end

  action.end_turn()

end

