_ENV = AIContext

---------- Priority Calculation ----------

------ Ships

-- targeting priority for ships is split into a base priority and an adjusted priority
-- the base priority is the combat value of a ship by itself, independent of whatever might be targeting it
-- it is used for both hostile and friendly ships, since the same criteria that make a ship an appealing 
-- target also indicate that the ship is some combination of valuable and vulnerable and needs defending.
function CombatAI_Method:base_ship_priority(ship)
  local attack_value = calc_attack_value(ship) + self:boarding_attack_value(ship)
  return attack_value/(1 + calc_defence_value(ship))
end

-- the adjusted ship priority is the priority of the a ship as a target for the given weapon
-- it is obtained from the base ship priority adjusted for range & distance, attenuation, etc
function CombatAI_Method:get_adjusted_ship_priority(target, weapon)
  local priority = self:base_ship_priority(target)

  local hit_chance = weapon:hit_chance(target) or 100
  priority *= (hit_chance/100)

  local to_target in range_fun
  local dist = to_target(weapon.object, target)

  if(weapon.is_beam)
    local range = weapon:range(dist)
    priority *= beam_attenuate(range)
  end

  if weapon.is_launcher
    local launcher_range = weapon:launcher_range()
    local turns_to_target = dist/launcher_range

    -- calculate priority multiplier for two separate concerns regarding launchers
    local point_def_mult = 1 --each turn our trackers are in flight byond the initial launch exposes them to PD
    local range_mult = 1 --for fightercraft, having them in flight for more than one turn means lost attacks, even if PD is absent

    --if each turn we lose missiles according to the est. survival rate, then the actual damage to the target is that rate ^ turns in-flight.
    if weapon.is_fighter
      --Note that 2*(0.45) - 0.9 == 0, this is intentional. 0.9 is 1-turn roundtrip range, 0.45 is this-turn roundtrip range
      point_def_mult = max(0.1, self.estimated_tracker_survival)^max(2*turns_to_target - 0.9, 0) --fighters have to return home, as well

      --if the target is 2 turns away we can only attack every 1 out of 2 turns, same for 3,4,5... etc. so divide priority by turns to target.
      range_mult = 1/max(turns_to_target, 0.9) --at 0.9 we can do a full round trip and be ready to attack by the start of next turn. So closer than that doesn't matter
    else
      --all targets closer than 0.9 can be hit in the same turn, so don't scale priority closer than that.
      point_def_mult = max(0.1, self.estimated_tracker_survival)^max(turns_to_target - 0.9, 0)
    end

    -- apply the lowest of the two
    priority *= min(point_def_mult, range_mult)
  end

  -- don't overcommit too many missiles/bombers to a target, if we think we can overcome hostile PD
  if self.estimated_tracker_survival > 0
    local incoming_trackers = self.enroute_trackers[target]
    if incoming_trackers
      local incoming_damage = 0
      for _, tracker_data in pairs(incoming_trackers) do
        incoming_damage += tracker_data.damage_per * tracker_data.n
      end

      local structure = max(0.7*target.structure - target.structure_damage, 0)
      local saturation_damage = ( calc_defence_value(target) + structure )/self.estimated_tracker_survival
      priority *= saturate(2 - incoming_damage/saturation_damage)
    end
  end

  return priority
end

----- Trackers

-- like base ship priority, this is the "inherent" priority of a tracker.
-- For example, destroying fighers and bombers prevents them from making further attacks in the future, unlike missiles.
-- NOTE this function is PER TRACKER in the group
function CombatAI_Method:base_tracker_priority_each(tracker)
  if tracker.is_fighter
    --again, it's important that this is damage/defence akin to ship priorities so that they remain comparable
    return average_tracker_damage(tracker)/tracker.health_per
  end

  return 0
end

--the value of intercepting trackers to prevent damage is based on the value of whatever they're targeting
local function tracker_intercept_priority(tracker, ai_data)
  -- get the priority of the ship this tracker is attacking
  if is_ship(tracker.dest) and tracker.dest ~= tracker.src
    return ai_data.ship_priority[tracker.dest]
  end

  -- in case we want to protect our own trackers from intercept
  if is_tracking_object(tracker.dest) and tracker.dest.empire == current_side()
    local friendly_tracker = tracker.dest
    local inherent_priority = ai_data:base_tracker_priority_each(friendly_tracker)*friendly_tracker.n
    if is_ship(friendly_tracker.dest) and friendly_tracker.dest ~= friendly_tracker.src
      return max(ai_data.ship_priority[friendly_tracker.dest], inherent_priority)
    end
    return inherent_priority
  end

  return 0
end

-- gets the priority of a tracker as a target for the given weapon
function CombatAI_Method:get_tracker_priority(tracker, weapon)

  -- ignore trackers that are already targeted by another tracker
  -- important to prevent every interceptor being launched against a single bomber group when there are many
  local already_targeted = 0
  local incoming_trackers = self.enroute_trackers[tracker]
  if incoming_trackers
    for _, tracker_data in pairs(incoming_trackers) do
      -- account for the fact that some trackers may take more than one hit to destroy
      -- while assuming that one hit can destroy at most one tracker
      already_targeted += min(tracker_data.damage_per/tracker.health_per, 1) * tracker_data.n
    end
  end
  --already_targeted = floor(already_targeted)

  local tracker_n = max(tracker.n - already_targeted, 0)

  -- inherent priority
  local inherent_priority = self:base_tracker_priority_each(tracker) * tracker_n

  -- intercept priority
  local intercept_priority = tracker_intercept_priority(tracker, self)

  -- apply a multipler that considers the amount of damage we are preventing
  -- versus the amount of damage to ships we could do with this weapon
  -- square relative damage to make it more heavily weighted, 
  -- but don't square the tracker count since it's unreliable
  local weapon_damage = average_weapon_damage(weapon)^2 * weapon:potential_shots()
  local tracker_damage = average_tracker_damage(tracker)^2 * tracker_n
  intercept_priority *= tracker_damage/weapon_damage 

  -- other modifiers
  local priority = max(inherent_priority, intercept_priority)

  local hit_chance = weapon:hit_chance(tracker) or 100
  priority *= (hit_chance/100)


  --there are none now, but lets assume there might be long range anti-tracker missiles
  if weapon.is_launcher
    local to_target in range_fun
    local dist = to_target(weapon.object, tracker)
    local launcher_range = weapon:launcher_range()
    local turns_to_target = dist/launcher_range

    local point_def_mult = 1 --each turn our trackers are in flight byond the initial launch exposes them to PD
    local range_mult = 1 --even if PD is absent, having fightercraft in flight for more than one turn means lost attacks (unlike missiles)

    --if each turn we lose missiles according to the est. survival rate, then the actual damage to the target is that rate ^ turns in-flight.
    if weapon.is_fighter
      --Note that 2*(0.45) - 0.9 == 0, this is intentional
      point_def_mult = max(0.1, self.estimated_tracker_survival)^max(2*turns_to_target - 0.9, 0) --fighters have to return home, as well

      --if the target is 2 turns away we can only attack every 1 out of 2 turns, same for 3,4,5... etc. so divide priority by turns to target.
      range_mult = 1/max(turns_to_target, 0.9) --at 0.9 we can do a full round trip and be ready to attack by the start of next turn. So closer than that doesn't matter
    else
      --all targets closer than 0.9 can be hit in the same turn, so don't scale priority closer than that.
      point_def_mult = max(0.1, self.estimated_tracker_survival)^max(turns_to_target - 0.9, 0)
    end

    priority *= min(point_def_mult, range_mult)

    -- if the tracker is a craft, check that we can make it there in time
    -- so that we don't end up chasing it all the way to dock
    if weapon.is_fighter and tracker.is_fighter
      local total_move_time in motion_function
      local tracker_range = tracker.speed*total_move_time
      local turns_to_dest = to_target(tracker, tracker.dest)/tracker_range

      if turns_to_target > ceil(turns_to_dest)
        priority *= 0.5
      end
    end

  end

  return priority
end

----- Launchers

-- Assigning priorities to hostile launchers allows the AI to make use of autofire
-- It is also very important for interceptors, since they may not always be able to "fire"
-- every turn, it can make a big difference to wait until a dangerous tracker is launched 
-- instead of wasting turns attacking some worthless target that is in combat now

-- very much like base_tracker_priority_each()
local function base_launcher_priority_each(launcher)
  if launcher.info.type == 'fighter'
    return average_weapon_damage(launcher)/launcher.info.health
  end

  return 0
end

-- this is very similar to get_tracker_priority(), except we are considering a tracker not yet launched
function CombatAI_Method:get_launcher_priority(launcher, weapon, intercept_priority)

  --don't want all our interceptors/PD reserved for the same launcher
  --pretty much the same calculation as for trackers
  local already_reserved = 0
  local reserved_weapons = self.reserved_weapons[launcher]
  if reserved_weapons
    for _, reserved_data in pairs(reserved_weapons) do
      already_reserved += min(reserved_data.damage_per/(launcher.info.health or 6), 1) * reserved_data.n
    end
  end
  --already_reserved = floor(already_reserved)

  -- the number of trackers the launcher can launch
  -- we are considering what they might do NEXT turn, so don't use useable_weapons()
  -- it doesn't matter that they fired this turn
  local launcher_n = max(alive_weapons(launcher) - launcher.missing - already_reserved, 0)

  -- inherent priority
  local inherent_priority = base_launcher_priority_each(launcher) * launcher_n

  -- intercept priority
  -- as an approximation, use our most valuable ship as the hypothetical target for this launcher
  intercept_priority = intercept_priority or self.highest_friendly_priority

  -- square relative damage to make it more heavily weighted, 
  -- but don't square the tracker count since it's unreliable
  local weapon_damage = average_weapon_damage(weapon)^2 * useable_weapons(weapon)
  local launcher_damage = average_weapon_damage(launcher)^2 * launcher_n
  intercept_priority *= launcher_damage/weapon_damage

  local priority = max(inherent_priority, intercept_priority)

  --there are none now, but lets assume there might be long range anti-tracker missiles
  if weapon.is_launcher
    --this is a little different than the calculation in get_tracker_priority(), for an important reason: 
    --here we are dealing with a hypothetical tracker that might launch NEXT turn
    --we make the assumption that this tracker will move in our general direction, so we subtract
    --it's launcher_range when determining turns_to_target, thus reducing the range penalty.
    --also don't care about dist < launcher_range
    local to_target in range_fun
    local dist = to_target(weapon.object, launcher.object)
    local launcher_range = weapon:launcher_range()
    local fighter_range = weapon:launcher_range()
    local turns_to_target = max(dist - launcher_range, 0)/fighter_range

    local point_def_mult = 1 --each turn our trackers are in flight byond the initial launch exposes them to PD
    local range_mult = 1 --even if PD is absent, having fightercraft in flight for more than one turn means lost attacks (unlike missiles)

    --if each turn we lose missiles according to the est. survival rate, then the actual damage to the target is that rate ^ turns in-flight.
    if weapon.is_fighter
      --Note that 2*(0.45) - 0.9 == 0, this is intentional
      point_def_mult = max(0.1, self.estimated_tracker_survival)^max(2*turns_to_target - 0.9, 0) --fighters have to return home, as well

      --if the target is 2 turns away we can only attack every 1 out of 2 turns, same for 3,4,5... etc. so divide priority by turns to target.
      range_mult = 1/max(turns_to_target, 0.9) --at 0.9 we can do a full round trip and be ready to attack by the start of next turn. So closer than that doesn't matter
    else
      --all targets closer than 0.9 can be hit in the same turn, so don't scale priority closer than that.
      point_def_mult = max(0.1, self.estimated_tracker_survival)^max(turns_to_target - 0.9, 0)
    end

    priority *= min(point_def_mult, range_mult)
  end

  --print(weapon.object.name, weapon.part, launcher.object.name, launcher.part, "inherent", inherent_priority, "intercept", intercept_priority)

  return priority
end

-- used when deciding to reserve PD for assault shuttles
function CombatAI_Method:get_assault_shuttle_priority(boarder, weapon)
  local shuttle_health_per = 6 --hardcoded in boarding_actions.lua

  --don't want all our interceptors/PD reserved for the same boarder
  --pretty much the same calculation as for trackers
  local already_reserved = 0
  local reserved_weapons = self.reserved_weapons[boarder]
  if reserved_weapons
    for _, reserved_data in pairs(reserved_weapons) do
      already_reserved += min(reserved_data.damage_per/shuttle_health_per, 1) * reserved_data.n * 0.5
    end
  end
  already_reserved = floor(already_reserved)

  --le sigh
  local base_n=0
  for s in systems(boarder)
    if s.parent_type=='Assault Shuttles'
      base_n+=3
    end
  end

  return self:boarding_attack_value(boarder) * max(base_n - already_reserved, 0)/base_n
end

----- Estimated point-defence survival rate

-- evaluate how hazardous the enemy fleet's PD is to our trackers
-- returns a value that reprents the fraction of our trackers we expect to survive each turn
function evaluate_hostile_point_defence(ai_data)

  -- count the total number of trackers we can launch
  -- and figure out our average tracker health
  -- TODO median tracker health might be a better metric
  local tracker_health,tracker_count = 0,0
  local point_defence = 0

  for ship in current_sides_ships() do
    for w in weapons(ship) do
      if w.is_launcher
        local n = w:potential_shots()
        local health = w.info.health or 6
        tracker_health += n*sqrt(health)
        tracker_count += n
      end
    end
  end

  if tracker_count == 0
    return 0 --no trackers
  end
  tracker_health /= tracker_count

  -- rough estimate of the number of trackers the enemy fleet could kill
  for obj in objects() do
    if ai_data:is_hostile(obj) --really annoying how enemies() can't take current_side() as an argument...
      if is_ship(obj)
        local ship = obj
        for w in weapons(ship) do
          if not w.cant_attack_trackers
            --factor in PD fired on their turn, but not too much
            local usable = useable_weapons(w)
            local alive = max(alive_weapons(w) - w.missing, 0)
            local n = (usable + 2*alive)/3

            local damage = average_weapon_damage(w)
            local hits = min(damage/tracker_health, 1) * n

            --miss chance fudge factor
            if w.is_pd or w.is_fighter
              hits *= 0.8
            else
              hits *= 0.5
            end

            point_defence += hits
          end
        end
      elseif is_tracking_object(obj)
        --no real way to tell if a tracking object is an 'interceptor', hopefully this is good enough
        if obj.is_fighter and is_tracking_object(obj.dest)
          for _, wname in pairs(obj.weapons) do
            local damage = WEAPONS[wname.parent_type].damage_fun.average
            point_defence += min(damage/tracker_health, 1) * obj.n
          end
        end
      end
    end
  end

  --estimated tracker survival rate
  local rval = tracker_count > 0 and max(tracker_count-point_defence, 0)/tracker_count or 0
  --print("ESTIMATED TRACKER SURVIVAL", current_side().name, tracker_count, point_defence, rval )
  return rval
end