_ENV = AIContext

----- Retreating

function CombatAI_Method:ship_combat_power(ship)
  local attack_value = calc_attack_value(ship) + self:boarding_attack_value(ship)
  return attack_value*(1+calc_defence_value(ship))
end

function CombatAI_Method:tracker_combat_power(tracker)
  local damage = average_tracker_damage(tracker)
  return (damage*tracker.health_per*2) * tracker.n
end

-- used for deciding whether or not to retreat ships
function evaluate_fleet_power(ai_data)
  local allied_power = 0
  local hostile_power = 0

  for obj in objects() do
    if ai_data:is_hostile(obj)
      hostile_power += ai_data.combat_power[obj]
    else
      allied_power += ai_data.combat_power[obj]
    end
  end

  return allied_power, hostile_power
end

function CombatAI_Method:evaluate_fleet_power()
  local allied_power, hostile_power = evaluate_fleet_power(self)
  self.allied_combat_power = allied_power
  self.hostile_combat_power = hostile_power
  self.outmatched = (allied_power == 0 and hostile_power == 0) and 0 or hostile_power/allied_power
end

-- how close are we to being destroyed?
-- return a value between 0 and 1
local function damage_assessment(ship)

  local shield = 0
  if ship.max_shields > 0
    local shield_damage = 0
    for i,s in pairs(SHIELD_NAMES) do
      shield_damage = max(shield_damage, ship[s])
    end
    shield = ship.max_shields - shield_damage
  end

  local armor = ship.armor - ship.armor_damage

  local max_structure = 0.7*ship.structure
  local structure = max(max_structure - ship.structure_damage, 0)
  
  -- since penetrating weapons do exist we don't want to rely too heavily on 
  -- undamaged shields/armor to tell us that we're not in danger
  -- so weigh them less when they are full, and more when they are damaged
  local armor_weight = ship.armor > 0 and (1 - 0.5*armor/ship.armor) or 0
  local shield_weight = ship.max_shields > 0 and (1 - 0.5*shield/ship.max_shields) or 0
  local structure_weight = max(1, armor_weight + shield_weight) --structure is always weighted at least equally to armor+shields

  local total = structure_weight*(structure/max_structure)
  total += shield_weight > 0 and shield_weight*(shield/ship.max_shields) or 0
  total += armor_weight > 0 and armor_weight*(armor/ship.armor) or 0

  local totalweight = shield_weight + armor_weight + structure_weight

  return saturate(1 - total/totalweight)
end

local function is_disarmed(ship)
  for w in weapons(ship)
   if alive_weapons(w) > 0
    return false
   end
  end
  return true
end

-- individual ship retreat decisions
function CombatAI_Method:wants_to_retreat(ship, general_retreat)
  if not can_retreat(ship)
    return false
  end

  if ship.empire == get_harpy_empire()
    return false --space harpies don't retreat
  end

  -- if a general retreat has been ordered, try to retreat all ships on the same turn
  -- that means ships that can retreat faster wait until the slower ships can retreat
  if general_retreat and ship.retreat_time >= general_retreat
    return true
  end

  local severely_damaged = damage_assessment(ship) --value between 0 and 1
  --print("DAMAGE ASSESSMENT", ship.name, severely_damaged)

  if is_disarmed(ship) and (severely_damaged > 0.1 or self.outmatched > 1.5)
    print(ship.empire.name, ship.name, "retreating because disarmed")
    return true
  end

  --possibly decide to withdraw armed ships if they are really severely damaged
  --or if removing them will not swing the balance of power
  if severely_damaged > 0.5
    print(ship.empire.name, ship.name, "retreating because severely damaged", severely_damaged)
    return true
  end

  return false
end

function CombatAI_Method:do_retreat(ship)

  action.retreat(ship)

  -- continue to fire weapons while retreating, unless we plan to cloak
  if not ship.cloaked -- if we're already cloaked, stay that way
    if (not ship.cant_cloak) and (is_disarmed(ship) or damage_assessment(ship) > 0.25)
      action.cloak {ship}
    else
      local temp_reserve_list = {}
      while (self:attack_as_able(ship, temp_reserve_list)) do end
    end
  end

end

function CombatAI_Method:general_retreat_threshold()
  local immobilized_power = 0

  for ship in current_sides_ships() do
    if ship.is_planet 
      return 5 --if we are defending a planet, try extra hard to hold it
    elseif ship.cant_move
      immobilized_power += self.combat_power[ship]
    end
  end

  return 3 + 2*(immobilized_power/self.allied_combat_power)
end

-- return the number of turns left until all ships retreat
function general_retreat()
  local rval = 0
  for ship in current_sides_ships() do
    if can_retreat(ship)
      if ship.retreating
        rval = max(rval, ship.retreating)
      else
        rval = max(rval, ship.retreat_time)
      end
    end
  end
  return rval
end